#include "llvm/Module.h"
#include "llvm/DerivedTypes.h"
#include "llvm/Instructions.h"
#include "llvm/Constants.h"
#include "llvm/BasicBlock.h"
#include "llvm/Pass.h"
#include "llvm/Support/CFG.h"
#include "llvm/Support/CallSite.h"
#include "llvm/IntrinsicInst.h"
#include "llvm/LinkAllVMCore.h"
#include "llvm/CodeGen/DwarfWriter.h"
#include "llvm/Analysis/DebugInfo.h"

#include "verify-fixrace.h"

using namespace std;
using namespace llvm;

namespace defens {
	struct verify_fixrace_stub: public BasicBlockPass {
		static char ID; // Pass identification
		verify_fixrace_stub(): BasicBlockPass(&ID) {}            
		virtual bool doInitialization(Module &M);
		virtual bool doFinalization(Module &M);
		virtual bool doFinalization(Function &F);
		virtual bool runOnBasicBlock(BasicBlock &BB);        
		virtual bool doInitialization(Function &F) {
			return BasicBlockPass::doInitialization(F);
		}
		bool checkFunctionName(Instruction *Inst, string &FuncName);
		int getFixOpInfo(Instruction *Inst, struct FixOpInfo *Info);
		bool DFSForward(Instruction *Inst, string &FixID, int FixType, bool IsFirst);
		bool DFSBackward(Instruction *Inst, string &FixID, int FixType, bool IsFirst);
    };

	bool verify_fixrace_stub::doInitialization(Module &M) {
		return true;
    }

	bool verify_fixrace_stub::doFinalization(Module &M) {
		return true;
	}

	bool verify_fixrace_stub::doFinalization(Function &F) {
		return true;
	}

	bool verify_fixrace_stub::checkFunctionName(Instruction *Inst, string &FuncName) {
		Function *F;
		
		if (Inst->getOpcode() == Instruction::Call) {
			F = dyn_cast<CallInst>(Inst)->getCalledFunction();
			if (F && F->getName() == FuncName) {
				return true;
			}
		}

		return false;
	}

	int verify_fixrace_stub::getFixOpInfo(Instruction *Inst, struct FixOpInfo *Info) {
		Constant *CT;
		CallSite CS;
		CallSite::arg_iterator AI;
		ostringstream FileName;
		string Temp;
		string FuncName = "printf";
		bool Matched = false;

		if (!checkFunctionName(Inst, FuncName))
			return -1;

		Info->FixID = "";
		Info->FixType = Info->FixOpType = -1;
		CS = CallSite::get(Inst);
        	if (CS.getInstruction() && CS.arg_size() == 1) {
        		AI = CS.arg_begin();
			if (isa<Constant>(AI)) {
				CT = dyn_cast<Constant>(AI)->getOperand(0);
				FileName << *CT;
				Temp = FileName.str();
				if (Temp.find("__FIXRACE__") == string::npos)
					return -1;

				// cerr << Temp << endl;
				istringstream FileNameChecked(Temp.c_str());
				while (FileNameChecked >> Temp) {
					if (Temp.find("__FIXRACE__") != string::npos) {
						Matched = true;
						continue;
					}
					
					if (Matched == true) {
						if (Info->FixID == "") {
							Info->FixID = Temp;
							continue;
						}
						
						if (Info->FixType == -1) {
							if (Temp == "CRITICAL") {
								Info->FixType = CRITICAL;
							} else if (Temp == "ATOMIC") {
								Info->FixType = ATOMIC;
							} else if (Temp == "EXISTING_CRITICAL") {
								Info->FixType = EXISTING_CRITICAL;
							} else {
								Info->FixType = UNKNOWN;
								return -1;
							}
							continue;
						} 

						if (Info->FixOpType == -1) {
							if (Temp.find("START") == 0) {
								Info->FixOpType = START;
								return 0;
							} else if (Temp.find("END") == 0) {
								Info->FixOpType = END;
								return 0;
							} else {
								Info->FixOpType = -1;
								return -1;
							}
						}
					}
				}
			}
		}

		return -1;
	}

	bool verify_fixrace_stub::DFSBackward(Instruction *Inst, string &FixID, int FixType, bool IsFirst) {
		struct FixOpInfo TargetInfo;
		BasicBlock *BB = Inst->getParent();
		Instruction *MatchedInst = NULL;
		Instruction *EndInst = &BB->back();
		bool Matched = false;

		cerr << "Backward Visit BB: " << BB->getName() << endl;

		// Return true here, any problem?
		if (VisitedBBs.find(BB) != VisitedBBs.end())
			return true;

		// Search current BB.
		do {
			BasicBlock::iterator I = Inst;
			if (IsFirst && I == BB->begin())
				break;
			else if (IsFirst) {
				cerr << "IsFirst Current Inst: " << *I;
				cerr << "IsFirst First Inst: " << *BB->begin();
				I--;
			}
	  		do {
				if (getFixOpInfo(I, &TargetInfo) == 0) {
					if (TargetInfo.FixID == FixID && TargetInfo.FixType == FixType && TargetInfo.FixOpType == START) {
						MatchedInst = I;
						Matched = true;
						break;
					} else if (TargetInfo.FixID == FixID && TargetInfo.FixType == FixType) {
						cerr << "Warn: END not alternative in BB: " << BB->getName() << endl;
						return false;
					}
				}		
				I--;
	  		} while (I != BB->begin());
		} while (0);
	
		return false;
	}


	bool verify_fixrace_stub::runOnBasicBlock(BasicBlock &BB) {
		struct FixOpInfo Info;
		BasicBlock::iterator I;
		
		for (BasicBlock::iterator IB = BB.begin(), IE = BB.end(); IB != IE; IB++) {
			if (getFixOpInfo(IB, &Info) == 0) {
				if (Info.FixOpType == END) {				
					cerr << "Inst: " << *IB;
					VisitedBBs.clear();
					if (DFSBackward(IB, Info.FixID, Info.FixType, true) == false)
						cerr << "Fail to match START finally!" << endl;
					else
						cerr << "Succeed to match START finally" << endl;				
				}
			}
		}
		
		return true;
	}
	
    char verify_fixrace_stub::ID = 0;
    static RegisterPass<verify_fixrace_stub> X("verify_fixrace_stub", "verify fixrace stub");
}

